#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# TODO The code with changed to use the v1 API. This is not tested and most likely this
# code will fail.
# TODO The metadata should not be set via API, but added to the uploads as nomad.json.

from typing import List
import requests
import re
import subprocess
from urllib import parse as urllib_parse
import os
import tarfile
import threading
import time
import typing
import io
import re
import uuid
import json
import numpy as np
import ase
import bs4
import matid

from nomad import atomutils, config, client, processing as proc
from nomad.client import api, upload_file


class DbUpdater:
    '''
    Automatically synchronizes nomad it with a given database. It creates a list of paths
    to mainfiles in nomad and compares it with paths in the external database. The missing
    paths in nomad will then be downloaded from the external database and subsequently
    uploaded to nomad. The downloaded files are by default saved in '/nomad/fairdi/external'.
    '''

    def __init__(self, *args, auth: client.Auth, **kwargs):
        self.db_name = 'aflowlib'
        self.root_url = 'http://aflowlib.duke.edu/AFLOWDATA/LIB1_LIB'
        self.local_path = '/nomad/fairdi/external'
        self.dbfile = None
        self.nomadfile = None
        self.outfile = None
        self.do_download = False
        self.do_upload = False
        self.do_publish = False
        self.cleanup = False
        self.parallel = 2
        self.max_depth = None
        self.target_file = None
        self.uids: List[str] = []
        self._set(**kwargs)
        self.auth = auth

    def _set(self, **kwargs):
        for key, val in kwargs.items():
            key = key.lower()
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                raise KeyError('Invalid key %s' % key)

        # create directory to save data
        subdir = ''
        if self.db_name.lower() == 'aflowlib':
            subdir = os.path.basename(self.root_url.strip('/'))

        dbpath = os.path.join(self.local_path, self.db_name)
        self._local_path = os.path.join(dbpath, subdir)
        if not os.path.isdir(dbpath):
            os.mkdir(dbpath)
        if not os.path.isdir(self._local_path):
            os.mkdir(self._local_path)

        # set target file based on database
        if self.db_name.lower() == 'aflowlib':
            if self.target_file is None:
                self.target_file = 'vasprun.xml.relax2.xz' if re.match(
                    r'.+/LIB\d+_LIB/?', self.root_url) else 'OUTCAR.static.xz'
        else:
            raise NotImplementedError('%s not yet supported.' % self.db_name)

        self._session = requests.Session()

    def _get_paths(self, root: str) -> typing.List[str]:
        response = self._session.get(root, verify=False)
        if not response.ok:
            response.raise_for_status()

        paths = []
        for url in re.findall('<a href="([^"]+)">', str(response.content)):
            if url in response.url:
                continue
            paths.append(os.path.join(response.url, url.lstrip('./')))
        return paths

    def _is_valid_file(self, path: str) -> bool:
        ok = True
        if '?' in path:
            ok = False
        return ok

    def _to_namesafe(self, path: str) -> str:
        path = path.strip('/')
        return re.sub(r'[^\w\d-]', '_', path)

    def _read_from_file(self, filename):
        print('Reading from file %s' % filename)
        data = []
        with open(filename) as f:
            line = f.readline()
            while line:
                data_i = line.strip().split()

                if len(data_i) == 1:
                    data.append(data_i[0])

                else:
                    info = data_i[1] in ['T', 'True', '1', 'Y']
                    data.append((data_i[0], info))

                line = f.readline()
        return data

    def _write_to_file(self, data: typing.List, filename: str):
        with open(filename, 'w') as f:
            for i in range(len(data)):
                if isinstance(data[i], str):
                    f.write('%s\n' % data[i])
                else:
                    f.write('%s %s \n' % (data[i][0], data[i][1]))

    def get_db_list(self):
        if self.dbfile is not None and os.path.isfile(self.dbfile):
            self.db_files = self._read_from_file(self.dbfile)
        else:
            print('Generating list from %s' % self.root_url)
            self.db_files = []
            todo = self._get_paths(self.root_url)
            while len(todo) > 0:
                cur = todo.pop(0)
                depth = urllib_parse.urlparse(cur).path.rstrip('/').count('/')
                if self.target_file in cur and self.max_depth is None:
                    # it makes the assumption that the the mainfiles are located at the
                    # same level
                    self.max_depth = depth - 1

                if self.max_depth is not None and depth > self.max_depth:
                    pass

                elif self.max_depth is not None and depth == self.max_depth:
                    self.db_files.append(cur)

                else:
                    try:
                        add = self._get_paths(cur)
                    except Exception:
                        add = []
                    add = [path for path in add if self._is_valid_file(path)]
                    todo = add + todo

            if self.dbfile is not None:
                self._write_to_file(self.db_files, self.dbfile)

    def get_nomad_list(self):
        if self.nomadfile is not None and os.path.isfile(self.nomadfile):
            self.nomad_files = self._read_from_file(self.nomadfile)
        else:
            print('Generating NOMAD list')
            if self.db_name.lower() == 'aflowlib':
                servers = ['LIB%d_LIB' % n for n in range(1, 10)] + ['ICSD_WEB']
                paths = [s for s in servers if s in self.root_url]
                paths = paths if paths else servers
                # main_author: Stefano Curtarolo
                query = dict(
                    main_author_id='81b96683-7170-49d7-8c4e-e9f34906b3ea',
                    paths=paths)

            self.nomad_files = []
            page_after_value = None
            while True:
                response = api.post('entries/query', data=json.dumps({
                    'owner': 'all',
                    'query': query,
                    'aggregations': {
                        'mainfiles': {
                            'terms': {
                                'quantity': 'mainfile',
                                'pagination': {
                                    'page_size': 1000,
                                    'page_after_value': page_after_value
                                }
                            }
                        }
                    }
                }))
                assert response.status_code == 200
                aggregation = response.json()['aggregations']['mainfiles']['terms']
                page_after_value = aggregation['pagination']['next_page_after_value']
                for bucket in aggregation['data']:
                    self.nomad_files.append(bucket['value'])

                if len(aggregation['data']) < 1000 or page_after_value is None:
                    break

            if self.nomadfile is not None:
                self._write_to_file(self.nomad_files, self.nomadfile)

    def compare_lists(self):
        '''
        Identify the difference between the nomad list and db list
        '''
        def reduce_list(ilist: typing.List[str]):
            olist = []
            for e in ilist:
                p = urllib_parse.urlparse(e).path.strip('/')
                olist.append(os.path.join(*p.split('/')[1:self.max_depth]))
            olist = list(set(olist))
            olist.sort()
            return olist

        print('Identifying differences')
        self.max_depth = 10 if self.max_depth is None else self.max_depth
        db = reduce_list(self.db_files)
        nomad = reduce_list(self.nomad_files)
        ns = set(nomad)
        self.update_list = [i for i in db if i not in ns]

        ds = set(db)
        in_nomad = [i for i in nomad if i not in ds]
        if len(in_nomad) > 0:
            fn = 'in_nomad.txt'
            print('Warning: Some NOMAD entries not found in db.')
            print('See %s for list.' % fn)
            self._write_to_file(in_nomad, fn)

        # add the root back
        u = urllib_parse.urlparse(self.root_url)
        up = u.path.strip('/').split('/')[0]
        root = '%s://%s/%s' % (u.scheme, u.netloc, up)
        self.update_list = [os.path.join(root, e) for e in self.update_list]
        self.is_updated_list = [False] * len(self.update_list)
        print('Found %d entries to be added in NOMAD' % len(self.update_list))

        if self.outfile is not None:
            data = [self.update_list[i] for i in range(len(self.update_list))]
            self._write_to_file(data, self.outfile)

    def _get_files(self, path: str) -> typing.Tuple[str, float]:
        def is_dir(path: str) -> bool:
            path = path.strip()
            if path[-1] == '/' and self.root_url in path:
                return True
            return False

        def download(path: str, iodir: str) -> float:
            files = self._get_paths(path)
            files = [f for f in files if self._is_valid_file(f)]
            size = 0.0
            for f in files:
                if is_dir(f):
                    _, s = self._get_files(f)
                    size += s
                    continue

                res = self._session.get(f, stream=True)
                fn = res.url.split('/')[-1]
                fn = os.path.join(iodir, fn)
                try:
                    with open(fn, 'wb') as fb:
                        fb.write(res.content)
                except Exception:
                    continue

                size += os.path.getsize(fn)
            return size

        def get_download_size(dirname: str) -> float:
            complete = False
            files = os.listdir(dirname)

            if self.db_name.lower() == 'aflowlib':
                complete = self.target_file in files
            else:
                complete = True

            size = 0.0
            if not complete:
                self._cleanup([os.path.join(dirname, f) for f in files])
            else:
                for f in files:
                    size += os.path.getsize(os.path.join(dirname, f))

            return size

        dirname = urllib_parse.urlparse(path).path
        dirname = self._to_namesafe(dirname)
        dirname = os.path.join(self._local_path, dirname)

        if os.path.isdir(dirname):
            size = get_download_size(dirname)
            if size == 0.0:
                size = download(path, dirname)

        else:
            try:
                os.mkdir(dirname)
                size = download(path, dirname)
            except Exception:
                size = 0.0

        return dirname, size

    def _make_name(self, dirs: typing.List[str]) -> typing.Tuple[str, str]:
        # name will be first and last entries
        d1 = self._to_namesafe(dirs[0].lstrip(self._local_path))
        d2 = self._to_namesafe(dirs[-1].lstrip(self._local_path))

        tarname = '%s-%s' % (d1, d2)
        uploadname = '%s_%s' % (self.db_name.upper(), tarname)
        tarname = os.path.join(self._local_path, '%s.tar' % tarname)

        return tarname, uploadname

    def _cleanup(self, ilist: typing.Union[str, typing.List[str]]):
        if isinstance(ilist, str):
            ilist = [ilist]
        for name in ilist:
            subprocess.Popen(['rm', '-rf', name])

    # TODO This should not be set via API, but added as nomad.json to the upload!
    # def get_payload(self, uid: int) -> typing.Dict[str, typing.Any]:
    #     if self.db_name == 'aflowlib':
    #         return dict(
    #             operation='publish',
    #             metadata=dict(
    #                 with_embargo=False,
    #                 comment='',
    #                 references=[
    #                     'http://www.sciencedirect.com/science/article/pii/S0927025614003322',
    #                     'http://aflowlib.org',
    #                     'http://www.sciencedirect.com/science/article/pii/S0927025612000687'],
    #                 coauthors=[
    #                     'f409d859-2639-4f82-b198-85e1c7c62f8b',
    #                     '580f7036-97b8-42b1-a9e6-815058bcac72',
    #                     '4c8d767d-335f-4ccd-9459-d0152b2026bc',
    #                     '7effd16a-a65c-4d95-b692-652488d94146',
    #                     'd59b3610-6335-4ad8-aca0-24a905de3a25',
    #                     '3df68fed-6ca0-4bf9-b2b1-d6be71e18f72',
    #                     'f7b540eb-a266-4379-9611-911ed8e3630e',
    #                     '7b3fe468-0011-4ba8-bd53-1ee65acda114',
    #                     '48e7a028-1a41-440d-9986-38540f5079c9',
    #                     'd63d07e6-ccc8-4eac-82af-d841627b6c53',
    #                     '9c308f66-eed1-4996-b33e-af78fb4944c7',
    #                     'd2621bc7-c45a-4d35-9dc1-5c05fa8326cb',
    #                     'ecba0e68-65ee-4b40-8fbf-a42714b1072b',
    #                     '81b96683-7170-49d7-8c4e-e9f34906b3ea'],
    #                 shared_with=[]))
    #     else:
    #         raise NotImplementedError('%s not yet supported.' % self.db_name)

    def publish(self, uids=None):
        def is_done_upload(uid: int) -> bool:
            response = api.get(f'uploads/{uid}', auth=self.auth)
            assert response.status_code == 200
            return response.json()['data']['process_status'] in proc.ProcessStatus.STATUSES_NOT_PROCESSING

        print('Publishing')
        uids = self.uids if uids is None else uids
        for uid in uids:
            if is_done_upload(uid):
                response = api.post(f'uploads/{uid}/action/publish', auth=self.auth)
                assert response.status_code == 200

    def upload(self, file_path: str, upload_name: str) -> int:
        uid = upload_file(os.path.abspath(file_path), self.auth, local_path=True, upload_name=upload_name)
        assert uid is not None
        return uid

    def _download_proc(self, plist: typing.List[str]):
        def tar_files(dirs: typing.List[str], tarname: str):
            if os.path.isfile(tarname):
                return

            try:
                with tarfile.open(tarname, 'w') as f:
                    for d in dirs:
                        files = os.listdir(d)
                        for fn in files:
                            f.add(os.path.join(d, fn))

            except Exception as e:
                os.remove(tarname)
                print('Error writing tar file %s. %s' % (tarname, e))

        def get_status_upload(uploadname: str) -> typing.Tuple[str, str]:
            response = api.get(f'uploads', params=dict(name=uploadname), auth=self.auth)
            assert response.status_code == 200
            response_json = response.json()
            assert len(response_json['data']) <= 1

            if len(response_json['data']) == 0:
                return None, None

            upload = response_json['data'][0]
            if upload['published']:
                return 'published', upload['upload_id']

            return 'uploaded', upload['upload_id']

        size = 0.0
        max_zip_size = config.max_upload_size
        dirs = []
        for i in range(len(plist)):
            d, s = self._get_files(self.update_list[plist[i]])
            if not self.do_upload:
                continue

            size += s
            dirs.append(d)
            if size > max_zip_size or i == (len(plist) - 1):
                tarname, uploadname = self._make_name(dirs)
                status, uid = get_status_upload(uploadname)
                if status == 'published':
                    continue
                if status != 'uploaded':
                    tar_files(dirs, tarname)
                    uid = upload_file(tarname, auth=self.auth, upload_name=uploadname, local_path=True)
                    assert uid is not None
                if self.do_publish:
                    self.publish([uid])
                self.uids.append(uid)
                if self.cleanup:
                    self._cleanup(dirs)
                    self._cleanup(tarname)
                size = 0.0
                dirs = []

    def download(self):
        '''
        Download files from database.
        '''
        print('Downloading from %s' % self.root_url)
        s = time.time()
        plist = [[] for i in range(self.parallel)]
        cur = 0
        for i in range(len(self.update_list)):
            if self.is_updated_list[i]:
                continue
            plist[cur % self.parallel].append(i)
            cur += 1

        if self.parallel > 1:
            procs = []
            for i in range(self.parallel):
                p = threading.Thread(target=self._download_proc, args=(plist[i],))
                procs.append(p)
            [p.start() for p in procs]
            [p.join() for p in procs]

        else:
            self._download_proc(plist[0])

        print('Time for download and upload (s)', time.time() - s)

    def get_list_to_download(self):
        '''
        Generate lists of files from database and from nomad and returns the difference.
        '''
        if self.outfile is not None and os.path.isfile(self.outfile):
            self.update_list = []
            self.is_updated_list = []
            for list_from_file in self._read_from_file(self.outfile):
                if isinstance(list_from_file, str):
                    self.update_list.append(list_from_file)
                    self.is_updated_list.append(False)
                else:
                    self.update_list.append(list_from_file[0])
                    self.is_updated_list.append(list_from_file[1])

        else:
            if self.parallel > 1:
                procs = []
                procs.append(threading.Thread(target=self.get_db_list))
                procs.append(threading.Thread(target=self.get_nomad_list))
                [p.start() for p in procs]
                [p.join() for p in procs]
            else:
                self.get_db_list()
                self.get_nomad_list()
            self.compare_lists()

    def update(self):
        self.get_list_to_download()
        if self.do_download:
            self.download()


def write_prototype_data_file(aflow_prototypes: dict, filepath) -> None:
    '''Writes the prototype data file in a compressed format to a python
    module.

    Args:
        aflow_prototypes
    '''
    class NoIndent(object):
        def __init__(self, value):
            self.value = value

    class NoIndentEncoder(json.JSONEncoder):
        '''A custom JSON encoder that can pretty-print objects wrapped in the
        NoIndent class.
        '''
        def __init__(self, *args, **kwargs):
            super(NoIndentEncoder, self).__init__(*args, **kwargs)
            self.kwargs = dict(kwargs)
            del self.kwargs['indent']
            self._replacement_map = {}

        def default(self, o):  # pylint: disable=E0202
            if isinstance(o, NoIndent):
                key = uuid.uuid4().hex
                self._replacement_map[key] = json.dumps(o.value, **self.kwargs)
                return "@@%s@@" % (key,)
            else:
                return super(NoIndentEncoder, self).default(o)

        def encode(self, o):
            result = super(NoIndentEncoder, self).encode(o)
            for k, v in self._replacement_map.items():
                result = result.replace('"@@%s@@"' % (k,), v)
            return result

    prototype_dict = aflow_prototypes["prototypes_by_spacegroup"]
    for prototypes in prototype_dict.values():
        for prototype in prototypes:
            # Save the information back in a prettified form
            prototype["atom_positions"] = NoIndent(prototype["atom_positions"])
            prototype["atom_labels"] = NoIndent(prototype["atom_labels"])
            prototype["lattice_vectors"] = NoIndent(prototype["lattice_vectors"])
            try:
                prototype["normalized_wyckoff_matid"] = NoIndent(prototype["normalized_wyckoff_matid"])
            except KeyError:
                pass

    # Save the updated data
    with io.open(filepath, "w", encoding="utf8") as f:
        json_dump = json.dumps(aflow_prototypes, ensure_ascii=False, indent=4, sort_keys=True, cls=NoIndentEncoder)
        json_dump = re.sub(r"\"(-?\d+(?:[\.,]\d+)?)\"", r'\1', json_dump)  # Removes quotes around numbers
        f.write("# -*- coding: utf-8 -*-\naflow_prototypes = {}\n".format(json_dump))


def update_prototypes(ctx, filepath, matches_only):

    if matches_only:
        from nomad.aflow_prototypes import aflow_prototypes
    else:
        # The basic AFLOW prototype data is available in a Javascript file. Here we
        # retrieve it and read only the prototype list from it.
        prototypes_file_url = 'http://aflowlib.org/CrystalDatabase/js/table_sort.js'
        r = requests.get(prototypes_file_url, allow_redirects=True)
        datastring = r.content.decode("utf-8")
        datastring = datastring.split('];')[0]
        datastring = datastring.split('= [')[1]
        data = json.loads('[' + datastring + ']')

        newdictarray = []
        n_prototypes = 0
        n_missing = 0
        for protodict in data:
            n_prototypes += 1
            newdict = {}

            # Make prototype plaintext
            prototype = bs4.BeautifulSoup(protodict["Prototype"], "html5lib").getText()

            # Add to new dictionary
            newdict['Notes'] = protodict['Notes']
            newdict['Prototype'] = prototype
            newdict['Space Group Symbol'] = protodict['Space Group Symbol']
            newdict['Space Group Number'] = protodict['Space Group Number']
            newdict['Pearsons Symbol'] = protodict['Pearson Symbol']
            newdict['Strukturbericht Designation'] = protodict['Strukturbericht Designation']
            newdict['aflow_prototype_id'] = protodict['AFLOW Prototype']
            newdict['aflow_prototype_url'] = 'http://www.aflowlib.org/CrystalDatabase/' + protodict['href'][2:]

            # Download cif or poscar if possible make ASE ase.Atoms object if possible
            # to obtain labels, positions, cell
            cifurl = 'http://www.aflowlib.org/CrystalDatabase/CIF/' + protodict['href'][2:-5] + '.cif'
            r = requests.get(cifurl, allow_redirects=True)
            cif_str = r.content.decode("utf-8")
            cif_file = io.StringIO()
            cif_file.write(cif_str)
            cif_file.seek(0)
            try:
                atoms = ase.io.read(cif_file, format='cif')
            except Exception:
                print("Error in getting prototype structure from CIF: {}", format(cifurl))
                # Then try to get structure from POSCAR
                try:
                    poscarurl = 'http://www.aflowlib.org/CrystalDatabase/POSCAR/' + protodict['href'][2:-5] + '.poscar'
                    r = requests.get(poscarurl, allow_redirects=True)
                    poscar_str = r.content.decode("utf-8")
                    poscar_file = io.StringIO()
                    poscar_file.write(poscar_str)
                    poscar_file.seek(0)
                    atoms = ase.io.read(poscar_file, format='vasp')
                except Exception:
                    print("Error in getting prototype structure from POSCAR: {}".format(poscarurl))
                    print("Could not read prototype structure from CIF or POSCAR file for prototype: {}, {}, ".format(prototype, newdict['aflow_prototype_url']))
                    n_missing += 1
                    continue

            atom_positions = atoms.get_positions()
            atom_labels = atoms.get_chemical_symbols()
            cell = atoms.get_cell()

            newdict['lattice_vectors'] = cell.tolist()
            newdict['atom_positions'] = atom_positions.tolist()
            newdict['atom_labels'] = atom_labels
            newdictarray.append(newdict)

            print("Processed: {}".format(len(newdictarray)))

        # Sort prototype dictionaries by spacegroup and make dictionary
        structure_types_by_spacegroup = {}
        for i_sg in range(1, 231):
            protos_sg = []
            for newdict in newdictarray:
                if newdict['Space Group Number'] == i_sg:
                    protos_sg.append(newdict)
            structure_types_by_spacegroup[i_sg] = protos_sg

        # Wrap in a dictionary that can hold other data, e.g. the symmemtry tolerance parameter.
        aflow_prototypes = {
            "prototypes_by_spacegroup": structure_types_by_spacegroup
        }
        print(
            "Extracted latest AFLOW prototypes online. Total number of "
            "successfully fetched prototypes: {}, missing: {}"
            .format(n_prototypes, n_missing)
        )

    # Update matches
    n_prototypes = 0
    n_failed = 0
    n_unmatched = 0
    prototype_dict = aflow_prototypes["prototypes_by_spacegroup"]

    for aflow_spg_number, prototypes in prototype_dict.items():
        n_prototypes += len(prototypes)
        for prototype in prototypes:

            # Read prototype structure
            pos = np.array(prototype["atom_positions"])
            labels = prototype["atom_labels"]
            cell = np.array(prototype["lattice_vectors"])
            atoms = ase.Atoms(
                symbols=labels,
                positions=pos,
                cell=cell,
                pbc=True
            )

            # Try to first see if the space group can be matched with the one in AFLOW
            try:
                symm = matid.SymmetryAnalyzer(atoms, config.normalize.prototype_symmetry_tolerance)
                spg_number = symm.get_space_group_number()
                wyckoff_matid = symm.get_wyckoff_letters_conventional()
                norm_system = symm.get_conventional_system()
            except Exception:
                n_failed += 1
            else:
                # If the space group is matched, add the MatID normalized Wyckoff
                # letters to the data.
                if spg_number == aflow_spg_number:
                    atomic_numbers = norm_system.get_atomic_numbers()
                    normalized_wyckoff_matid = atomutils.get_normalized_wyckoff(atomic_numbers, wyckoff_matid)
                    prototype["normalized_wyckoff_matid"] = normalized_wyckoff_matid
                else:
                    n_unmatched += 1
    print(
        "Updated matches in AFLOW prototype library. Total number of "
        "prototypes: {}, unmatched: {}, failed: {}"
        .format(n_prototypes, n_unmatched, n_failed)
    )

    # Write data file to the specified path
    write_prototype_data_file(aflow_prototypes, filepath)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation
from nomad.datamodel.metainfo import workflow


m_package = Package()


class x_castep_section_vibrational_frequencies(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_number_vibrational_frequencies = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of vibration frequenices
        ''')

    x_castep_vibrational_frequencies = Quantity(
        type=np.dtype(np.float64),
        shape=['x_castep_number_vibrational_frequencies'],
        description='''
        Vibration Frequenices (cm-1)
        ''')

    x_castep_vibrationl_frequencies_store = Quantity(
        type=str,
        shape=[],
        description='''
        Vibration Frequenices (cm-1)
        ''')

    x_castep_ir_store = Quantity(
        type=str,
        shape=[],
        description='''
        Irreducible representation in the Point Group
        ''')

    x_castep_ir = Quantity(
        type=str,
        shape=['x_castep_number_vibrational_frequencies'],
        description='''
        Irreducible representation in the Point Group
        ''')

    x_castep_raman_activity = Quantity(
        type=np.dtype(np.float64),
        shape=['x_castep_number_vibrational_frequencies'],
        description='''
        Raman activity (A**4/amu)
        ''')

    x_castep_raman_active = Quantity(
        type=str,
        shape=['x_castep_number_vibrational_frequencies'],
        description='''
        Raman activity (A**4/amu)
        ''')

    x_castep_n_iterations_phonons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of iterations in phonons
        ''')

    x_castep_raman_activity_store = Quantity(
        type=str,
        shape=[],
        description='''
        Raman activity (A**4/amu)
        ''')

    x_castep_ir_intensity = Quantity(
        type=np.dtype(np.float64),
        shape=['x_castep_number_vibrational_frequencies'],
        description='''
        IR intensities (D/A)**2/amu
        ''')

    x_castep_ir_intensity_store = Quantity(
        type=str,
        shape=[],
        description='''
        IR intensities (D/A)**2/amu
        ''')


class x_castep_section_band_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_band_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_castep_band_n_bands
        ''')

    x_castep_band_conv_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_band_conv_tolerance
        ''')

    x_castep_band_n_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_castep_band_n_iterations
        ''')

    x_castep_band_max_cg = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_band_max_cg
        ''')


class x_castep_section_core_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_core_spectra_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        x_castep_core_spectra_n_bands
        ''')

    x_castep_core_spectra_conv_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_core_spectra_conv_tolerance
        ''')


class x_castep_section_ts_scf_iteration(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_scf_ts_iteration_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy
        ''')

    x_castep_scf_ts_iteration_energy_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy change
        ''')

    x_castep_scf_ts_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy time
        ''')


class x_castep_section_tddft(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_tddft_iteration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Iteration number
        ''')

    x_castep_wall_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Iteration wall time (s)
        ''')

    x_castep_state_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        state number
        ''')

    x_castep_state_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        state energy
        ''')

    x_castep_state_energy_error = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        state energy error
        ''')

    x_castep_tddft_calculation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        calculation time
        ''')


class x_castep_section_atom_ionic_velocities(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_castep_section_atom_positions_optim(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_cell_angle_alpha_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle alpha
        ''')

    x_castep_cell_angle_beta_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle beta
        ''')

    x_castep_cell_angle_gamma_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle gamma
        ''')

    x_castep_cell_length_a_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        a unit cell edge length
        ''')

    x_castep_cell_length_b_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        b unit cell edge length
        ''')

    x_castep_cell_length_c_optim = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        c unit cell edge length
        ''')


class x_castep_section_atom_positions(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_cell_angle_alpha = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle alpha
        ''')

    x_castep_cell_angle_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle beta
        ''')

    x_castep_cell_angle_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Simulation cell angle gamma
        ''')

    x_castep_cell_length_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        a unit cell edge length
        ''')

    x_castep_cell_length_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        b unit cell edge length
        ''')

    x_castep_cell_length_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        c unit cell edge length
        ''')


class x_castep_section_cell_optim(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_cell_vector_optim = Quantity(
        type=str,
        shape=[],
        description='''
        Temporay storage for cell vectors
        ''')


class x_castep_section_cell(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_cell_vector = Quantity(
        type=str,
        shape=[],
        description='''
        Temporay storage for cell vectors
        ''')


class x_castep_section_collect_scf_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_castep_section_eigenvalues_1(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_eigenvalues_1 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storing eigenvalues
        ''')


class x_castep_section_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storing eigenvalues
        ''')


class x_castep_section_functional_definition(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_functional_type = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional definition in CASTEP convention
        ''')

    x_castep_functional_weight = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional weight in CASTEP convention
        ''')


class x_castep_section_functionals(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_functional_name = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional in CASTEP convention
        ''')


class x_castep_section_population_analysis(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_mulliken_atom_index = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_atom_index
        ''')

    x_castep_mulliken_atom = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_atom kind
        ''')

    x_castep_orbital_contributions = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_contributions
        ''')

    x_castep_orbital_s = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital s
        ''')

    x_castep_orbital_p = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital p
        ''')

    x_castep_orbital_d = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital d
        ''')

    x_castep_orbital_f = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_contribution_orbital f
        ''')

    x_castep_total_orbital = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_total_contribution
        ''')

    x_castep_mulliken_charge_store = Quantity(
        type=str,
        shape=[],
        description='''
        Mulliken_charges
        ''')

    x_castep_mulliken_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Mulliken_charges
        ''')


class x_castep_section_geom_optimisation_method(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_geometry_optim_method = Quantity(
        type=str,
        shape=[],
        description='''
        Determines optimisation method used
        ''')


class x_castep_section_optics_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_optics_n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        optics_number_of_bands
        ''')

    x_castep_optics_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        optics_band_convergence_tolerance
        ''')


class x_castep_section_tddft_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_tddft_n_excited_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of excited states
        ''')

    x_castep_tddft_n_states_forces = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of states for forces
        ''')

    x_castep_tddft_state_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        tolerance (eV)
        ''')

    x_castep_tddft_state_tolerance_window = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        tolerance window iterations
        ''')

    x_castep_tddft_max_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        max number iterations
        ''')

    x_castep_tddft_extra_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of extra states
        ''')

    x_castep_tddft_functional = Quantity(
        type=str,
        shape=[],
        description='''
        tddft functional
        ''')

    x_castep_tddft_method = Quantity(
        type=str,
        shape=[],
        description='''
        tddft method
        ''')

    x_castep_tddft_eigenmethod = Quantity(
        type=str,
        shape=[],
        description='''
        tddft eigenmethod
        ''')

    x_castep_tddft_approximation = Quantity(
        type=str,
        shape=[],
        description='''
        tddft approximation
        ''')

    x_castep_tddft_position_op = Quantity(
        type=str,
        shape=[],
        description='''
        tddft position operator
        ''')


class x_castep_section_phonons(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_phonon_method = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon calculation method
        ''')

    x_castep_DFPT_solver_method = Quantity(
        type=str,
        shape=[],
        description='''
        Phonon DFPT solver method
        ''')

    x_castep_phonon_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon calculation tolerance (eV/A**2)
        ''')

    x_castep_phonon_cycles = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon calculation cycles
        ''')

    x_castep_band_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Phonon band convergence tolerance window
        ''')


class x_castep_section_density_mixing_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_density_mixing_scheme = Quantity(
        type=str,
        shape=[],
        description='''
        density_mixing_scheme
        ''')

    x_castep_density_mixing_length = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density_mixing_scheme_length
        ''')

    x_castep_charge_density_mixing_amplitude = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge_density_mixing_amplitude
        ''')

    x_castep_cut_off_energy_for_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge_density_mixing_cut_off_energy_for_mixing (A)
        ''')


class x_castep_section_population_analysis_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_population_analysis_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Population_analysis_cutoff_(A)
        ''')


class x_castep_section_ts_scf(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_scf_ts_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy Total
        ''')

    x_castep_scf_ts_total_energy_free = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy Total free
        ''')

    x_castep_scf_ts_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        SCF_ts_energy T0 free
        ''')

    x_castep_section_ts_scf_iteration = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_scf_iteration'),
        repeats=True)


class x_castep_section_k_band(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)


class x_castep_section_k_points_1(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_k_points_1 = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing k points coordinates (fractional)
        ''')


class x_castep_section_k_points(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_k_points = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing k points coordinates (fractional)
        ''')


class x_castep_section_relativity_treatment(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_relativity_treatment_scf = Quantity(
        type=str,
        shape=[],
        description='''
        Relativity treatment in CASTEP convention
        ''')


class x_castep_section_scf_eigenvalues(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_scf_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')


class x_castep_section_SCF_iteration_frame(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_frame_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_store_t_md_frame
        ''')

    x_castep_SCF_frame_energy = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_iterations
        ''')

    x_castep_SCF_frame_energy_gain = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_iterations_gain
        ''')

    x_castep_frame_time_scf_iteration_wall_end = Quantity(
        type=str,
        shape=[],
        description='''
        energy_frame_wall_end_time
        ''')

    x_castep_frame_energy_free = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        energy_free
        ''')

    x_castep_frame_energy_total_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        energy_free_corrected_for_finite_basis
        ''')


class x_castep_section_scf_k_points(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_scf_k_points = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_castep_section_scf_eigenvalues = SubSection(
        sub_section=SectionProxy('x_castep_section_scf_eigenvalues'),
        repeats=True)


class x_castep_section_spin_number(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_spin_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')


class x_castep_section_stress_tensor(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_stress_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing stress tensor components
        ''')


class x_castep_section_time(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_calculation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        castep_calculation_time
        ''')

    x_castep_finalisation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        castep_finalisation_time
        ''')

    x_castep_initialisation_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        x_castep_Initialisation_time
        ''')


class x_castep_section_raman_tensor(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_store_raman_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged Raman susceptibility tensor
        ''')

    x_castep_raman_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        # TODO verify unit, original parser was ampere / amu
        unit='angstrom / unified_atomic_mass_unit',
        description='''
        Ramen tensor
        ''')


class x_castep_section_scf_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_energy_threshold = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Energy Threshold
        ''')

    x_castep_max_iter = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of maximum iterations steps
        ''')

    x_castep_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Smearing kind
        ''')

    x_castep_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Smearing width
        ''')


class x_castep_section_electronic_spectroscpy_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_theory_level = Quantity(
        type=str,
        shape=[],
        description='''
        Electronic spectroscopy parameters theory level
        ''')

    x_castep_spectroscopy_calculation = Quantity(
        type=str,
        shape=[],
        description='''
        Electronic spectroscopy parameters calculation type
        ''')

    x_castep_spec_max_iter = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of iterations
        ''')

    x_castep_spec_max_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of steps
        ''')

    x_castep_spec_max_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Max number of bands
        ''')

    x_castep_spec_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Tolerance (eV)
        ''')


class x_castep_section_md(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_md_energies = Quantity(
        type=str,
        shape=[],
        description='''
        md_energy_components
        ''')

    x_castep_md_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_temp(K)
        ''')

    x_castep_md_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_pressure
        ''')

    x_castep_md_cell_vectors = Quantity(
        type=str,
        shape=[],
        description='''
        md_cell_vectors
        ''')

    x_castep_md_cell_vectors_vel = Quantity(
        type=str,
        shape=[],
        description='''
        md_cell_vectors_velocities
        ''')

    x_castep_md_stress_tensor = Quantity(
        type=str,
        shape=[],
        description='''
        md_stress_tensor
        ''')

    x_castep_md_positions = Quantity(
        type=str,
        shape=[],
        description='''
        md_positions
        ''')

    x_castep_md_forces = Quantity(
        type=str,
        shape=[],
        description='''
        md_forces
        ''')

    x_castep_md_lab = Quantity(
        type=str,
        shape=[],
        description='''
        md_lables
        ''')

    x_castep_md_veloc = Quantity(
        type=str,
        shape=[],
        description='''
        md_veloc
        ''')


class x_castep_section_ts(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_total
        ''')

    x_castep_ts_cell_vectors = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_cell_vectors
        ''')

    x_castep_ts_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_positions
        ''')

    x_castep_ts_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_forces
        ''')

    x_castep_ts_lab = Quantity(
        type=str,
        shape=[],
        description='''
        ts_lables
        ''')


class x_castep_section_ts_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_ts_store
        ''')

    x_castep_ts_cell_vectors_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors
        ''')

    x_castep_ts_positions_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_positions_store
        ''')

    x_castep_ts_forces_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_forces_store
        ''')


class x_castep_section_ts_final_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_energy_final_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_ts
        ''')

    x_castep_ts_forces_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_forces_final
        ''')

    x_castep_ts_positions_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_positions_final
        ''')

    x_castep_ts_cell_vectors_final_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors_final
        ''')


class x_castep_section_ts_final(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_energy_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_final
        ''')

    x_castep_ts_cell_vectors_final = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_cell_vectors
        ''')

    x_castep_ts_positions_final = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_positions_final
        ''')

    x_castep_ts_forces_final = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_force_finals
        ''')


class x_castep_section_ts_product(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_energy_product = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_prod
        ''')

    x_castep_ts_forces_product = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_energy_forces_pro
        ''')

    x_castep_ts_positions_product = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        ts_energy_positions_pro
        ''')

    x_castep_ts_cell_vectors_product = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        ts_product_cell_vectors
        ''')


class x_castep_section_ts_product_store(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_energy_product_store = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_energy_components_pro
        ''')

    x_castep_ts_forces_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_forces_pro_store
        ''')

    x_castep_ts_positions_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_energy_positions_pro_store
        ''')

    x_castep_ts_cell_vectors_pro_store = Quantity(
        type=str,
        shape=[],
        description='''
        ts_cell_vectors_pro
        ''')


class x_castep_section_ts_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_ts_method = Quantity(
        type=str,
        shape=[],
        description='''
        ts_method
        ''')

    x_castep_ts_protocol = Quantity(
        type=str,
        shape=[],
        description='''
        ts_protocol
        ''')

    x_castep_ts_number_qst = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_qst_iterations
        ''')

    x_castep_ts_number_cg = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_number_of_cg_iterations
        ''')

    x_castep_ts_force_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_force_tolerance (eV/A)
        ''')

    x_castep_ts_displacement_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_displacement_tolerance (A)
        ''')


class x_castep_section_DFT_SEDC(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        correlation energy
        ''')

    x_castep_de_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dE/atom
        ''')

    x_castep_dfmax_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dfmax/atom
        ''')

    x_castep_structure_energy_corr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        structure energy correction
        ''')

    x_castep_PBC_image_inter_corr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        PBC image interaction corr.
        ''')

    x_castep_total_energy_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total energy correction
        ''')

    x_castep_total_fmax_correction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        correction F max ev/A
        ''')

    x_castep_shell = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        shell
        ''')

    x_castep_total_dispersion_corrected_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total_dispersion_corrected_free_energy
        ''')

    x_castep_disp_corrected_energy_total_T0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dispersion corrected zero point
        ''')


class x_castep_section_van_der_Waals_parameters(MSection):
    '''
    -
    '''

    m_def = Section(validate=False)

    x_castep_disp_method_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name type
        ''')

    x_castep_Parameter_d = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method G06
        ''')

    x_castep_Parameter_LAMBDA = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method OBS
        ''')

    x_castep_Parameter_n = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method OBS
        ''')

    x_castep_Parameter_s6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method G06
        ''')

    x_castep_Parameter_sR = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Parameter for dispersion method TS
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_atom_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Forces on the atoms as minus gradient of energy_total, including forces' unitary-
        transformation (rigid body) filtering and including constraints, if present. The
        derivatives with respect to displacements of the nuclei in the gradient are
        evaluated according to the coordinate system defined in coordinate_system. In
        addition, these forces are obtained by filtering out the unitary transformations
        (translations of the center of mass and rigid rotations of the whole system, when
        non periodic), atom_forces_raw for the unfiltered counterpart. Furthermore, forces
        due to constraints like fixed atoms, distances, angles, dihedrals, and so on, are
        here included (see atom_forces_raw for the unfiltered counterpart).
        ''')

    x_castep_basis_set_planewave_cutoff_iteration_0 = Quantity(
        type=str,
        shape=[],
        description='''
        cutoff at iteration 0 of geometry optimisation
        ''')

    x_castep_compiler = Quantity(
        type=str,
        shape=[],
        description='''
        Compiler name
        ''')

    x_castep_constants_reference = Quantity(
        type=str,
        shape=[],
        description='''
        Fundamental constant data source
        ''')

    x_castep_fft_library = Quantity(
        type=str,
        shape=[],
        description='''
        fft library name
        ''')

    x_castep_geom_converged = Quantity(
        type=str,
        shape=[],
        description='''
        CASTEP_geom_converged
        ''')

    x_castep_maths_library = Quantity(
        type=str,
        shape=[],
        description='''
        Maths library name
        ''')

    x_castep_program_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        Compilation date (string)
        ''')

    x_castep_program_compilation_time = Quantity(
        type=str,
        shape=[],
        description='''
        Compilation time (string)
        ''')

    x_castep_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        Run execution date (string)
        ''')

    x_castep_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        Run execution start time (string)
        ''')

    x_castep_crystal_point_group = Quantity(
        type=str,
        shape=[],
        description='''
        Point group of the crystal (Schoenflies notation)
        ''')

    x_castep_space_group = Quantity(
        type=str,
        shape=[],
        description='''
        Point space of the crystal
        ''')

    x_castep_ts_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_number
        ''')

    x_castep_ts_path_ts_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_final
        ''')

    x_castep_ts_path_product = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ts_path_pro
        ''')

    x_castep_store_atom_forces_band = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_castep_elec_methd = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing electronic structure method
        ''')

    x_castep_section_vibrational_frequencies = SubSection(
        sub_section=SectionProxy('x_castep_section_vibrational_frequencies'),
        repeats=True)

    x_castep_section_band_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_band_parameters'),
        repeats=True)

    x_castep_section_core_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_core_parameters'),
        repeats=True)

    x_castep_section_collect_scf_eigenvalues = SubSection(
        sub_section=SectionProxy('x_castep_section_collect_scf_eigenvalues'),
        repeats=True)

    x_castep_section_population_analysis = SubSection(
        sub_section=SectionProxy('x_castep_section_population_analysis'),
        repeats=True)

    x_castep_section_geom_optimisation_method = SubSection(
        sub_section=SectionProxy('x_castep_section_geom_optimisation_method'),
        repeats=True)

    x_castep_section_optics_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_optics_parameters'),
        repeats=True)

    x_castep_section_tddft_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_tddft_parameters'),
        repeats=True)

    x_castep_section_phonons = SubSection(
        sub_section=SectionProxy('x_castep_section_phonons'),
        repeats=True)

    x_castep_section_density_mixing_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_density_mixing_parameters'),
        repeats=True)

    x_castep_section_population_analysis_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_population_analysis_parameters'),
        repeats=True)

    x_castep_section_SCF_iteration_frame = SubSection(
        sub_section=SectionProxy('x_castep_section_SCF_iteration_frame'),
        repeats=True)

    x_castep_section_time = SubSection(
        sub_section=SectionProxy('x_castep_section_time'),
        repeats=True)

    x_castep_section_raman_tensor = SubSection(
        sub_section=SectionProxy('x_castep_section_raman_tensor'),
        repeats=True)

    x_castep_section_scf_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_scf_parameters'),
        repeats=True)

    x_castep_section_electronic_spectroscpy_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_electronic_spectroscpy_parameters'),
        repeats=True)

    x_castep_section_md = SubSection(
        sub_section=SectionProxy('x_castep_section_md'),
        repeats=True)

    x_castep_section_ts = SubSection(
        sub_section=SectionProxy('x_castep_section_ts'),
        repeats=True)

    x_castep_section_ts_store = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_store'),
        repeats=True)

    x_castep_section_ts_final_store = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_final_store'),
        repeats=True)

    x_castep_section_ts_final = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_final'),
        repeats=True)

    x_castep_section_ts_product = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_product'),
        repeats=True)

    x_castep_section_ts_product_store = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_product_store'),
        repeats=True)

    x_castep_section_ts_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_parameters'),
        repeats=True)

    x_castep_section_van_der_Waals_parameters = SubSection(
        sub_section=SectionProxy('x_castep_section_van_der_Waals_parameters'),
        repeats=True)


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_net_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Net charge of system
        ''')

    x_castep_number_of_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of bands
        ''')

    x_castep_number_of_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of electrons
        ''')

    x_castep_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Storing atomic positions in fractional coordinates
        ''')

    x_castep_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_cell_volume
        ''')

    x_castep_optimised_atom_labels = Quantity(
        type=str,
        shape=['number_of_atoms'],
        description='''
        Temporary storing atomic positions
        ''')

    x_castep_optimised_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Storing atomic optimised positions in fractional coordinates
        ''')

    x_castep_velocities_cell_vector = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        cell vector velocities
        ''')

    x_castep_store_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom labels
        ''')

    x_castep_store_atom_number = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atom labels
        ''')

    x_castep_store_atom_positions = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_castep_store_atom_ionic_velocities = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_castep_atom_ionic_velocities = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        Temporary storing atomic positions
        ''')

    x_castep_store_optimised_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_castep_store_optimised_atom_positions = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing atomic positions
        ''')

    x_castep_section_tddft = SubSection(
        sub_section=SectionProxy('x_castep_section_tddft'),
        repeats=True)

    x_castep_section_atom_ionic_velocities = SubSection(
        sub_section=SectionProxy('x_castep_section_atom_ionic_velocities'),
        repeats=True)

    x_castep_section_atom_positions_optim = SubSection(
        sub_section=SectionProxy('x_castep_section_atom_positions_optim'),
        repeats=True)

    x_castep_section_atom_positions = SubSection(
        sub_section=SectionProxy('x_castep_section_atom_positions'),
        repeats=True)

    x_castep_section_cell_optim = SubSection(
        sub_section=SectionProxy('x_castep_section_cell_optim'),
        repeats=True)

    x_castep_section_cell = SubSection(
        sub_section=SectionProxy('x_castep_section_cell'),
        repeats=True)


class BasisSetCellDependent(simulation.method.BasisSetCellDependent):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_basis_set_planewave_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing plane wave cutoff as string
        ''')

    x_castep_size_std_grid = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        size of standard grid (eV)
        ''')

    x_castep_size_fine_grid = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        sise of fine grid (1/A)
        ''')


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_enthalpy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_Enthalpy
        ''')

    x_castep_frequency = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_frequency (cm-1)
        ''')

    x_castep_improved_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        md_forces
        ''')

    x_castep_frame_time_0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_store_t_md_frame
        ''')

    x_castep_geom_iteration_index = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Index for number of iterations in geometry optimisation
        ''')

    x_castep_store_atom_forces = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storing converged atom forces (ev/A)
        ''')

    x_castep_initial_scf_iteration_wall_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Initial SCF iteration wall time
        ''')

    x_castep_total_dispersion_corrected_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Total electronic energy that includes dispersion energy computed with
        Disp_method_name not corrected for finite basis-set
        ''')

    x_castep_total_energy_corrected_for_finite_basis = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASTEP_total_energy_corrected_for_finite_basis
        ''')

    x_castep_total_energy_corrected_for_finite_basis_store = Quantity(
        type=str,
        shape=[],
        description='''
        CASTEP_total_energy_corrected_for_finite_basis_store
        ''')

    energy_total_scf_iteration_list = Quantity(
        type=np.dtype(np.float64),
        shape=[-1],
        description='''
        Total electronic energy calculated with XC_method_scf during the scf iterations is
        stored in a list
        ''')

    x_castep_section_ts_scf = SubSection(
        sub_section=SectionProxy('x_castep_section_ts_scf'),
        repeats=True)

    x_castep_section_scf_k_points = SubSection(
        sub_section=SectionProxy('x_castep_section_scf_k_points'),
        repeats=True)

    x_castep_section_spin_number = SubSection(
        sub_section=SectionProxy('x_castep_section_spin_number'),
        repeats=True)

    x_castep_section_stress_tensor = SubSection(
        sub_section=SectionProxy('x_castep_section_stress_tensor'),
        repeats=True)

    x_castep_section_DFT_SEDC = SubSection(
        sub_section=SectionProxy('x_castep_section_DFT_SEDC'),
        repeats=True)


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_functional_and_weight = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional+weight in CASTEP convention
        ''')

    x_castep_section_functional_definition = SubSection(
        sub_section=SectionProxy('x_castep_section_functional_definition'),
        repeats=True)

    x_castep_section_functionals = SubSection(
        sub_section=SectionProxy('x_castep_section_functionals'),
        repeats=True)

    x_castep_section_relativity_treatment = SubSection(
        sub_section=SectionProxy('x_castep_section_relativity_treatment'),
        repeats=True)


class MolecularDynamics(workflow.MolecularDynamics):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        thermostat_target_temperature(K)
        ''')

    x_castep_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        barostat_type
        ''')

    x_castep_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        thermostat_type
        ''')

    x_castep_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        thermostat_type
        ''')

    x_castep_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        barostat_tau
        ''')

    x_castep_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD_time_step (ps)
        ''')

    x_castep_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        MD_time_step_number
        ''')

    x_castep_frame_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        MD_pressure
        ''')

    x_castep_frame_energy_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        MD_scf_energy tolerance (eV)
        ''')

    x_castep_frame_eigen_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        MD_scf_eigen tolerance (eV)
        ''')


class GeometryOptimization(workflow.GeometryOptimization):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_geometry_stress_com_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        tolerance for stress components in geometry optimisation (GPa)
        ''')

    x_castep_max_number_of_steps = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number_of iterations geom_optim
        ''')


class BandStructure(simulation.calculation.BandStructure):

    m_def = Section(validate=False, extends_base_section=True)

    x_castep_k_path = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_castep_store_k_path = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_castep_store_k_label = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''')

    x_castep_section_eigenvalues_1 = SubSection(
        sub_section=SectionProxy('x_castep_section_eigenvalues_1'),
        repeats=True)

    x_castep_section_eigenvalues = SubSection(
        sub_section=SectionProxy('x_castep_section_eigenvalues'),
        repeats=True)

    x_castep_section_k_band = SubSection(
        sub_section=SectionProxy('x_castep_section_k_band'),
        repeats=True)

    x_castep_section_k_points_1 = SubSection(
        sub_section=SectionProxy('x_castep_section_k_points_1'),
        repeats=True)

    x_castep_section_k_points = SubSection(
        sub_section=SectionProxy('x_castep_section_k_points'),
        repeats=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os.path
import time

from .api import Auth


def upload_file(
        file_path: str, auth: Auth,
        upload_name: str = None, local_path: bool = False, publish: bool = False):
    '''
    Upload a file to nomad. Will print status information to stdout.

    Arguments:
        file_path: path to the file, absolute or relative to current working directory
        auth: mandatory Auth instance
        upload_name: optional name, default is the file_path's basename
        local_path: allows to process data without upload, requires client to be run on the server
        publish: automatically publish after successful processing

    Returns: The upload_id if successful or None if not.
    '''
    from nomad.processing import ProcessStatus
    from nomad.client import api
    if local_path:
        response = api.post(
            'uploads',
            params=dict(local_path=os.path.abspath(file_path), upload_name=upload_name),
            headers={'Accept': 'application/json'},
            auth=auth)
        print('process offline: %s' % file_path)
    else:
        with open(file_path, 'rb') as f:
            response = api.post(
                'uploads',
                params=dict(upload_name=upload_name), data=f,
                headers={'Accept': 'application/json'},
                auth=auth)
    if response.status_code != 200:
        print('Could not create upload: %s' % response.text)
        return None
    upload = response.json()['data']

    while upload['process_status'] not in [ProcessStatus.SUCCESS, ProcessStatus.FAILURE]:
        response = api.get(f'uploads/{upload["upload_id"]}/entries', auth=auth)
        response_json = response.json()

        upload = response_json['upload']
        total = response_json['pagination']['total']
        failures = response_json['processing_failed']
        successes = response_json['processing_successful']

        ret = '\n' if upload['process_status'] in (ProcessStatus.SUCCESS, ProcessStatus.FAILURE) else '\r'

        print(
            'status: %s; process: %s; parsing: %d/%d/%d                %s' %
            (upload['process_status'], upload['current_process'], successes, failures, total, ret), end='')

        time.sleep(1)

    if upload['process_status'] == ProcessStatus.FAILURE:
        print('There have been errors:')
        for error in upload['errors']:
            print('    %s' % error)
    elif publish:
        response = api.post(f'uploads/{upload["upload_id"]}/action/publish', auth=auth)
        if response.status_code != 200:
            print('Could not publish upload: %s' % response.text)
            return None

    return upload['upload_id']



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, Reference
)

from nomad.datamodel.metainfo import simulation
from nomad.metainfo.metainfo import JSON


m_package = Package()


class Run(simulation.run.Run):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_git_rev = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_process_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_psi4_psidatadir = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_memory = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        # unit='MiB',
        description='''
        ''')

    x_psi4_threads = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_psi4_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class System(simulation.system.System):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_molecular_point_group = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_full_point_group = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_molecular_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_rotational_constants = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        # unit='1/cm',
        description='''
        ''')

    x_psi4_nuclear_repulsion = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_multiplicity = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_nalpha = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_nbeta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class Scf(simulation.method.Scf):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_diis = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_mom = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_fractional_occupation = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_guess_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_integral_threshold = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class BasisSetAtomCentered(simulation.method.BasisSetAtomCentered):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_blend = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_n_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Gives the number of shell types used.
        ''')

    x_psi4_max_angular_momentum = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum angular momentum quantum number corresponding to the shells used.
        ''')

    x_psi4_n_cartesian_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_psi4_spherical_harmonics = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_n_ecp_primitives = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_psi4_n_ecp_core_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')


class Method(simulation.method.Method):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_scf_algorithm_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_diis = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_mom = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_fractional_occupation = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_psi4_guess_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_psi4_options = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_psi4_jk_matrices_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_psi4_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')


class DFT(simulation.method.DFT):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_molecular_quadrature = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        ''')


class x_psi4_root_information(MSection):
    m_def = Section(validate=False)

    x_psi4_root_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class Calculation(simulation.calculation.Calculation):
    m_def = Section(validate=False, extends_base_section=True)

    x_psi4_s2_expected = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_s2_observed = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_s_expected = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_s_observed = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_psi4_root_information = SubSection(sub_section=x_psi4_root_information.m_def, repeats=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation
from nomad.datamodel.metainfo import workflow


m_package = Package()


class x_charmm_mdin_input_output_files(MCategory):
    '''
    Parameters of mdin belonging to x_charmm_section_control_parameters.
    '''

    m_def = Category()


class x_charmm_mdin_control_parameters(MCategory):
    '''
    Parameters of mdin belonging to x_charmm_section_control_parameters.
    '''

    m_def = Category()


class x_charmm_mdin_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category()


class x_charmm_mdout_single_configuration_calculation(MCategory):
    '''
    Parameters of mdout belonging to section_single_configuration_calculation.
    '''

    m_def = Category()


class x_charmm_mdout_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category()


class x_charmm_mdout_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category()


class x_charmm_mdin_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category()


class x_charmm_section_input_output_files(MSection):
    '''
    Section to store input and output file names
    '''

    m_def = Section(validate=False)


class x_charmm_section_control_parameters(MSection):
    '''
    Section to store the input and output control parameters
    '''

    m_def = Section(validate=False)

    x_charmm_inout_file_structure = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input topology file.
        ''')

    x_charmm_inout_file_trajectory = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output trajectory file.
        ''')

    x_charmm_inout_file_traj_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output trajectory file.
        ''')

    x_charmm_inout_file_traj_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output file for velocities in the trajectory.
        ''')

    x_charmm_inout_file_traj_force = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output file for forces in the trajectory.
        ''')

    x_charmm_inout_file_output_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output coordinates file.
        ''')

    x_charmm_inout_file_out_coor_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output coordinates on log.
        ''')

    x_charmm_inout_file_output_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output velocities file.
        ''')

    x_charmm_inout_file_output_force = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output forces file.
        ''')

    x_charmm_inout_file_input_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input coordinates file.
        ''')

    x_charmm_inout_file_in_coor_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input coordinate on log file.
        ''')

    x_charmm_inout_file_input_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input velocities file.
        ''')

    x_charmm_inout_file_restart_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm restart coordinates file.
        ''')

    x_charmm_inout_file_restart_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm restart velocities file.
        ''')

    x_charmm_inout_file_rtf_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm RTF residue file.
        ''')

    x_charmm_inout_file_prm_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm PRM parameter file.
        ''')

    x_charmm_inout_file_cor_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm CRD coordinates file.
        ''')

    x_charmm_inout_file_stream = Quantity(
        type=str,
        shape=[],
        description='''
        charmm stream input/output.
        ''')

    x_charmm_inout_file_rtf_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm stream RTF input.
        ''')

    x_charmm_inout_file_par_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm stream parameter input.
        ''')

    x_charmm_inout_file_output_log = Quantity(
        type=str,
        shape=[],
        description='''
        charmm MD output log file.
        ''')

    x_charmm_inout_control_gaussian_option_is = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_crystal = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_crystal_type = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_a_length = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_b_length = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_c_length = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_alpha = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_beta = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_gamma = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_mini = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nstep = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_inbfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_step = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_prtmin = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tolfun = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tolgrd = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tolitr = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tolstp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tfreq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pcut = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ihbfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ncgcyc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nprint = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nonbond_option_flags = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_cutnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ctexnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ctonnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ctofnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_cgonnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_cgofnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_wmin = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_cdie = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_switch = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_vswitch = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_atoms = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_wrnmxd = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_e14fac = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_eps = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nbxmod = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bond_cutoff_distance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bond_cutoff_angle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bond_switching_on_distance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bond_switching_off_distance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bond_switching_on_angle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bond_switching_off_angle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_exclusions_due_to_distance_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_exclusions_due_to_angle_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_acceptor_antecedents = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_exclusions_due_to_duplications = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_deletions_due_to_best_option = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_deletions_due_to_duplications = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_deletions_due_to_fixed_atoms = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hbond_deletions_due_to_exclusion = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hydrogen_bonds_present = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_minimization_exit_status = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_dyna = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_akmastp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_firstt = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iseed = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iprfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ihtfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ieqfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iunrea = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iunwri = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iunos = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iuncrd = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iunvel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iunxyz = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_kunit = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nsavc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nsavv = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nsavx = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iscale = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iscvel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iasors = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_iasvel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ichecw = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ntrfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ilbfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_imgfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_isvfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_ncycle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_nsnos = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_teminc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tstruc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_finalt = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_twindh = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_twindl = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_time_step = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_random_num_gen_seeds = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_number_of_degrees_of_freedom = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_gaussian_option = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_velocities_assigned_at_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_seeds = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_shake_tolerance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_strt = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_rest = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_qref = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tref = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_hoover = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_reft = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_tmass = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pcons = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pmass = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_number_of_steps = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_steps_per_cycle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_initial_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_dielectric = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_excluded_species_or_groups = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_1_4_electrostatics_scale = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_velocity_rescale_temp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_velocity_reassignment_freq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_velocity_reassignment_temp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_dynamics = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_damping_coefficient_unit = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_temperature_coupling = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_coupling_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_target_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_oscillation_period = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_decay_time = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_langevin_piston_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pressure_control = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_particle_mesh_ewald = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pme_tolerance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pme_ewald_coefficient = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pme_interpolation_order = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pme_grid_dimensions = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_pme_maximum_grid_spacing = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_minimization = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_verlet_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_random_number_seed = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_structure_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_parameter_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_number_of_parameters = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_inout_control_parameters = Quantity(
        type=str,
        shape=['x_charmm_inout_control_number_of_parameters'],
        description='''
        charmm running environment and control parameters.
        ''')

    x_charmm_section_input_output_files = SubSection(
        sub_section=SectionProxy('x_charmm_section_input_output_files'),
        repeats=True)


class x_charmm_section_atom_to_atom_type_ref(MSection):
    '''
    Section to store atom label to atom type definition list
    '''

    m_def = Section(validate=False)

    x_charmm_atom_to_atom_type_ref = Quantity(
        type=np.dtype(np.int64),
        shape=['number_of_atoms_per_type'],
        description='''
        Reference to the atoms of each atom type.
        ''')


class x_charmm_section_single_configuration_calculation(MSection):
    '''
    section for gathering values for MD steps
    '''

    m_def = Section(validate=False)


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_charmm_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''')

    x_charmm_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''')

    x_charmm_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''')

    x_charmm_lattice_lengths = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Lattice dimensions in a vector. Vector includes [a, b, c] lengths.
        ''',)

    x_charmm_lattice_angles = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Angles of lattice vectors. Vector includes [alpha, beta, gamma] in degrees.
        ''',)

    x_charmm_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''')

    x_charmm_mdin_finline = Quantity(
        type=str,
        shape=[],
        description='''
        finline in mdin
        ''')

    x_charmm_traj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_traj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_traj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_traj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_traj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_traj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')


class MolecularDynamics(workflow.MolecularDynamics):

    m_def = Section(validate=False, extends_base_section=True)

    x_charmm_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''',)

    x_charmm_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''',)

    x_charmm_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''',)

    x_charmm_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''',)

    x_charmm_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''',)

    x_charmm_periodicity_type = Quantity(
        type=str,
        shape=[],
        description='''
        Periodic boundary condition type in the sampling (non-PBC or PBC).
        ''',)

    x_charmm_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''',)

    x_charmm_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''',)

    x_charmm_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''',)

    x_charmm_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''',)

    x_charmm_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''',)

    x_charmm_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''',)


class AtomParameters(simulation.method.AtomParameters):

    m_def = Section(validate=False, extends_base_section=True)

    x_charmm_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Atom name of an atom in topology definition.
        ''')

    x_charmm_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''')

    x_charmm_atom_element = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''')

    x_charmm_atom_type_element = Quantity(
        type=str,
        shape=[],
        description='''
        Element symbol of an atom type.
        ''')

    x_charmm_atom_type_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        van der Waals radius of an atom type.
        ''')

    number_of_atoms_per_type = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this type.
        ''')


class Interaction(simulation.method.Interaction):

    m_def = Section(validate=False, extends_base_section=True)

    x_charmm_interaction_atom_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''')

    x_charmm_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''')

    x_charmm_pair_interaction_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['x_charmm_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''')

    x_charmm_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''')

    x_charmm_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''')

    x_charmm_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''')

    x_charmm_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''')

    x_charmm_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['x_charmm_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_charmm_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''')

    x_charmm_parallel_task_nr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Program task no.
        ''')

    x_charmm_build_osarch = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build OS/ARCH
        ''')

    x_charmm_output_created_by_user = Quantity(
        type=str,
        shape=[],
        description='''
        Output file creator
        ''')

    x_charmm_most_severe_warning_level = Quantity(
        type=str,
        shape=[],
        description='''
        Highest charmm warning level in the run.
        ''')

    x_charmm_program_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build date
        ''')

    x_charmm_program_citation = Quantity(
        type=str,
        shape=[],
        description='''
        Program citations
        ''')

    x_charmm_program_copyright = Quantity(
        type=str,
        shape=[],
        description='''
        Program copyright
        ''')

    x_charmm_number_of_tasks = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of tasks in parallel program (MPI).
        ''')

    x_charmm_program_module_version = Quantity(
        type=str,
        shape=[],
        description='''
        charmm program module version.
        ''')

    x_charmm_program_license = Quantity(
        type=str,
        shape=[],
        description='''
        charmm program license.
        ''')

    x_charmm_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''')

    x_charmm_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''')

    x_charmm_program_working_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_program_execution_host = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_program_execution_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_program_module = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_mdin_header = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_mdin_wt = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_charmm_section_control_parameters = SubSection(
        sub_section=SectionProxy('x_charmm_section_control_parameters'),
        repeats=True)


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_charmm_section_single_configuration_calculation = SubSection(
        sub_section=SectionProxy('x_charmm_section_single_configuration_calculation'),
        repeats=True)
